/*

Copyright (C) 2003 Christian Kreibich <christian@whoop.org>.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies of the Software and its documentation and acknowledgment shall be
given in the documentation and software packages that this Software was
used.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <protocols/tcp/libnd_tcp.h>
#include "libnd_tcpstate.h"

#define LND_TCP_MSL         60
#define LND_TCP_TIMEOUT   5*60

struct lnd_tcp_conn
{
  /* Connection definition -- IP addr + port quadruple: */
  struct in_addr      ip_src;
  struct in_addr      ip_dst;
  guint16             th_sport;
  guint16             th_dport;

  /* TCP connection status in state machine: */
  LND_TCPState        state;

  /* Sequence numbers we keep track of. */
  guint32             syn;
  guint32             syn_ack;
  guint32             fin;
  guint32             fin_back;

  /* Timestamp at which connection is complete, and timing-out starts.
   * This is used for both error timeouts (after RSTs) and normal
   * timeouts in TIME_WAIT.
   */
  struct bpf_timeval  wait_ts;
  
  /* We time out during handshake and shutdown after 5 minutes. */
  struct bpf_timeval  timeout_ts;

  /* Arbitrary data storage hashtable. */
  GHashTable         *data;

  /* Various flags for our connection: */
  guchar              fin_acked:1;
  guchar              fin_back_acked:1;
  guchar              handshake_seen:1;
};


/* String table to represent states in text */
static const char *tcp_state_strings[] = {
  "ERROR",
  "LISTEN",
  "CLOSED_NORMAL",
  "RST_WAIT",
  "CLOSED_RST",
  "SYN_SENT",
  "SYN_ACK_SENT",
  "ESTABLISHED",
  "SHUTDOWN",
  "TIME_WAIT",
  "CLOSED_TIMEOUT"
};

/* GHashFunc to hash TCP connections: */
static guint
tcp_conn_hash(const LND_TCPConn *conn)
{
  return
    ((guint32) conn->th_sport ^ (guint32) conn->ip_src.s_addr)   ^
    ((guint32) conn->th_dport ^ (guint32) conn->ip_dst.s_addr);
}

/* GCompareFunc to compare our TCP Connections: */
static gint 
tcp_conn_compare(const LND_TCPConn *c1, const LND_TCPConn *c2)
{
  /* Terminated connections stay in the hashtable but
   * are never matched successfully.
   * FIXME: make this check only the connection currently stored in the hashtable.
   */
  if (libnd_tcpconn_is_dead(c1) ||
      libnd_tcpconn_is_dead(c2))
    {
      return FALSE;
    }

  /*
  printf("Comparing %u %u %u %u to %u %u %u %u\n",
	 c1->ip_src.s_addr, ntohs(c1->th_sport),
	 c1->ip_dst.s_addr, ntohs(c1->th_dport),
	 c2->ip_src.s_addr, ntohs(c2->th_sport),
	 c2->ip_dst.s_addr, ntohs(c2->th_dport));
  */ 

  /* Direct match: */
  if (c1->ip_src.s_addr == c2->ip_src.s_addr &&
      c1->ip_dst.s_addr == c2->ip_dst.s_addr &&
      c1->th_sport == c2->th_sport           &&
      c1->th_dport == c2->th_dport)
    {
      return TRUE;
    }
  
  /* Reverse direction match -- same flow! */
  if (c1->ip_src.s_addr == c2->ip_dst.s_addr &&
      c1->ip_dst.s_addr == c2->ip_src.s_addr &&
      c1->th_sport == c2->th_dport           &&
      c1->th_dport == c2->th_sport)
    {
      return TRUE;
    }
  
  return FALSE;
}

LND_TCPConn    *
libnd_tcpconn_new(const LND_Packet *packet)
{
  struct ip *iphdr;
  struct tcphdr *tcphdr;
  LND_TCPConn *tcpc;

  if (! (tcpc = g_new0(LND_TCPConn, 1)))
    return NULL;
  if (! libnd_tcp_get_headers(packet, &iphdr, &tcphdr))
    return NULL;
  
  tcpc->ip_src = iphdr->ip_src;
  tcpc->ip_dst = iphdr->ip_dst;
  tcpc->th_sport = tcphdr->th_sport;
  tcpc->th_dport = tcphdr->th_dport;

  tcpc->state = LND_TCP_LISTEN;

  tcpc->data = g_hash_table_new(g_str_hash, g_str_equal);

  return tcpc;
}


void            
libnd_tcpconn_free(LND_TCPConn *tcpc)
{
  g_hash_table_destroy(tcpc->data);
  g_free(tcpc);
}


static void
tcpconn_state_goto_shutdown(LND_TCPConn *tcpc,
			    const LND_Packet *packet,
			    struct ip *iphdr,
			    struct tcphdr *tcphdr)
{
  gboolean forward_flow = (iphdr->ip_src.s_addr == tcpc->ip_src.s_addr);
  
  if (forward_flow)
    tcpc->fin = ntohl(tcphdr->th_seq) + libnd_tcp_get_payload_length(iphdr, tcphdr);
  else
    tcpc->fin_back = ntohl(tcphdr->th_seq) + libnd_tcp_get_payload_length(iphdr, tcphdr);
  
  tcpc->state = LND_TCP_SHUTDOWN;
  tcpc->timeout_ts = packet->ph.ts;
}


void            
libnd_tcpconn_state_update(LND_TCPConn *tcpc,
			   const LND_Packet *packet)
{
  struct ip *iphdr;
  struct tcphdr *tcphdr;  
  gboolean forward_flow;
  struct bpf_timeval diff;

  if (!tcpc || !packet)
    return;
  if (! libnd_tcp_get_headers(packet, &iphdr, &tcphdr))
    return;
  
  forward_flow = (iphdr->ip_src.s_addr == tcpc->ip_src.s_addr);
  
  if ((tcphdr->th_flags & TH_RST) &&
      tcpc->state != LND_TCP_RST_WAIT)      
    {
      tcpc->state   = LND_TCP_RST_WAIT;
      tcpc->wait_ts = packet->ph.ts;
      return;
    }
  
  switch (tcpc->state)
    {
    case LND_TCP_LISTEN:
      /* If we get a SYN, we jump straigt to ESTABLISHED if
       * we also have an ACK. We assume we missed the initial SYN.
       * If there's no ACK, we start a timeout and go to SYN_SENT.
       */
      if ( (tcphdr->th_flags & TH_SYN))
	{
	  if ( (tcphdr->th_flags & TH_ACK))
	    {
	      tcpc->state = LND_TCP_ESTABLISHED;

	      if (forward_flow)
		{
		  tcpc->ip_src = iphdr->ip_dst;
		  tcpc->ip_dst = iphdr->ip_src;
		  tcpc->th_sport = tcphdr->th_dport;
		  tcpc->th_dport = tcphdr->th_sport;
		}
	    }
	  else
	    {
	      tcpc->state = LND_TCP_SYN_SENT;
	      tcpc->syn = ntohl(tcphdr->th_seq);
	      tcpc->timeout_ts = packet->ph.ts;
	    }
	  break;
	}

      /* If we don't see the connection establishment, we
       * assume we missed it and jump directly to ESTABLISHED.
       */
      tcpc->state = LND_TCP_ESTABLISHED;
      break;
      
    case LND_TCP_SYN_SENT:
      /* In SYN_SENT, first make sure that the timeout didn't
       * expire. If so, we go to CLOSED_TIMEOUT.
       */
      pcapnav_timeval_sub(&packet->ph.ts, &tcpc->timeout_ts, &diff);
      
      if (diff.tv_sec >= LND_TCP_TIMEOUT)
	{
	  tcpc->state = LND_TCP_CLOSED_TIMEOUT;
	  break;
	}

      /* Update our timeout */
      tcpc->timeout_ts = packet->ph.ts;

      /* If we SYN-ACK of the right byte going in opposite direction,
       * we're SYN_ACK_SENT and we still have to check for the pure ACK.
       */
      if ( (tcphdr->th_flags & TH_SYN) &&
	   (tcphdr->th_flags & TH_ACK) &&
	   (ntohl(tcphdr->th_ack) == tcpc->syn + 1) &&
	   ! forward_flow)
	{
	  tcpc->syn_ack = ntohl(tcphdr->th_seq);
	  tcpc->state = LND_TCP_SYN_ACK_SENT;
	  break;
	}

      /* And if we FIN, we attempt a connection teardown,
       * including a timeout in case this goes wrong.
       */
      if ( (tcphdr->th_flags & TH_FIN))
	tcpconn_state_goto_shutdown(tcpc, packet, iphdr, tcphdr);
      break;

    case LND_TCP_SYN_ACK_SENT:
      /* In SYN_ACK_SENT, first make sure that the timeout didn't
       * expire. If so, we go to CLOSED_TIMEOUT.
       */
      pcapnav_timeval_sub(&packet->ph.ts, &tcpc->timeout_ts, &diff);
      
      if (diff.tv_sec >= LND_TCP_TIMEOUT)
	{
	  tcpc->state = LND_TCP_CLOSED_TIMEOUT;
	  break;
	}

      /* Update our timeout */
      tcpc->timeout_ts = packet->ph.ts;

      if ( (tcphdr->th_flags == TH_ACK) &&
	   (ntohl(tcphdr->th_ack) == tcpc->syn_ack + 1))
	{
	  tcpc->state = LND_TCP_ESTABLISHED;
	  tcpc->handshake_seen = TRUE;
	}
      break;
      
    case LND_TCP_ESTABLISHED:
      /* We go from ESTABLISHED to SHUTDOWN if we see a
       * FIN and start a timeout in case the shutdown goes
       * wrong.
       */
      if ( (tcphdr->th_flags & TH_FIN))
	tcpconn_state_goto_shutdown(tcpc, packet, iphdr, tcphdr);
      break;

    case LND_TCP_SHUTDOWN:
      /* First, check whether the shutdown times out, in which
       * case we are CLOSED_TIMEOUT. Otherwise, track the shutdown
       * and make sure we see FIN plus ACK in both directions.
       * If that is the case, we enter TIME_WAIT.
       */
      pcapnav_timeval_sub(&packet->ph.ts, &tcpc->timeout_ts, &diff);
      
      if (diff.tv_sec >= LND_TCP_TIMEOUT)
	{
	  tcpc->state = LND_TCP_CLOSED_TIMEOUT;
	  break;
	}

      /* Update the timeout */
      tcpc->timeout_ts = packet->ph.ts;

      if ( (tcphdr->th_flags & TH_FIN))
	tcpconn_state_goto_shutdown(tcpc, packet, iphdr, tcphdr);

      if (forward_flow)
	{
	  if ( tcpc->fin_back &&
	       (tcphdr->th_flags & TH_ACK) &&
	       ntohl(tcphdr->th_ack) == tcpc->fin_back + 1)
	    tcpc->fin_back_acked = TRUE;
	}
      else
	{
	  if ( tcpc->fin &&
	       (tcphdr->th_flags & TH_ACK) &&
	       ntohl(tcphdr->th_ack) == tcpc->fin + 1)
	    tcpc->fin_acked = TRUE;
	}
	 
      if (tcpc->fin_acked && tcpc->fin_back_acked)
	{
	  tcpc->state = LND_TCP_TIME_WAIT;
	  tcpc->wait_ts = packet->ph.ts;
	}
      break;

    case LND_TCP_TIME_WAIT:
      /* In TIME_WAIT, we wait 2MSLs, after which
       * we simply move to CLOSED_NORMAL.
       */
      pcapnav_timeval_sub(&packet->ph.ts, &tcpc->wait_ts, &diff);
      
      if (diff.tv_sec >= 2 * LND_TCP_MSL)
	tcpc->state = LND_TCP_CLOSED_NORMAL;	
      break;

    case LND_TCP_RST_WAIT:
      /* After a RST we expect nothing and wait for the
       * timeout. Any packets arriving during this period
       * are assumed to have been in flight when we saw
       * the RST, and are ignored. After the timeout we
       * end at CLOSED_RST.
       */
      pcapnav_timeval_sub(&packet->ph.ts, &tcpc->wait_ts, &diff);
      
      if (diff.tv_sec >= LND_TCP_MSL)
	tcpc->state = LND_TCP_CLOSED_RST;
      break;
      
    case LND_TCP_ERROR:
    default:
      D(("Warning -- catch-all case reached in TCP state machine.\n"));
      break;
    }
}


gboolean        
libnd_tcpconn_handshake_seen(LND_TCPConn *tcpc)
{
  if (!tcpc)
    return FALSE;

  return tcpc->handshake_seen;
}


gboolean        
libnd_tcpconn_teardown_seen(LND_TCPConn *tcpc)
{
  if (! tcpc)
    return FALSE;
  
  return ( (tcpc->state == LND_TCP_CLOSED_NORMAL) ||
	   (tcpc->state == LND_TCP_CLOSED_RST)    ||
	   (tcpc->state == LND_TCP_TIME_WAIT)     ||
	   (tcpc->state == LND_TCP_RST_WAIT));
}


gboolean
libnd_tcpconn_is_dead(const LND_TCPConn *tcpc)
{
  if (! tcpc)
    return FALSE;
  
  return ( (tcpc->state == LND_TCP_CLOSED_NORMAL) ||
	   (tcpc->state == LND_TCP_CLOSED_RST)    ||
	   (tcpc->state == LND_TCP_CLOSED_TIMEOUT));
}


void
libnd_tcpconn_get_src(const LND_TCPConn *tcpc,
		      struct in_addr *ip_src,
		      guint16 *th_sport)
{
  if (!tcpc)
    return;

  if (ip_src)
    *ip_src = tcpc->ip_src;
  if (th_sport)
    *th_sport = tcpc->th_sport;
}


void
libnd_tcpconn_get_dst(const LND_TCPConn *tcpc,
		      struct in_addr *ip_dst,
		      guint16 *th_dport)
{
  if (!tcpc)
    return;

  if (ip_dst)
    *ip_dst = tcpc->ip_dst;
  if (th_dport)
    *th_dport = tcpc->th_dport;
}


void            
libnd_tcpconn_to_string(const LND_TCPConn *tcpc,
			char *str, int strlen)
{
  char ip_dst[16];

  if (!tcpc || !str || strlen == 0)
    return;

  g_snprintf(ip_dst, 16, "%s", inet_ntoa(tcpc->ip_dst));
  g_snprintf(str, strlen, "%s:%u - %s:%u",
	     inet_ntoa(tcpc->ip_src), ntohs(tcpc->th_sport),
	     ip_dst, ntohs(tcpc->th_dport));
}


LND_TCPState
libnd_tcpconn_state_get(const LND_TCPConn *tcpc)
{
  if (!tcpc)
    return LND_TCP_ERROR;

  return tcpc->state;
}


const char     *
libnd_tcpconn_state_get_string(LND_TCPState state)
{
  if (state <= LND_TCP_TIME_WAIT)
    return tcp_state_strings[state];

  return "undefined";
}

void            
libnd_tcpconn_data_set(LND_TCPConn *tcpc,
		       const char *key,
		       void *data)
{
  if (!tcpc || !key || !*key)
    return;
  
  g_hash_table_insert(tcpc->data, (gpointer) key, (gpointer) data);
}


void           *
libnd_tcpconn_data_get(const LND_TCPConn *tcpc,
		       const char *key)
{
  if (!tcpc || !key || !*key)
    return NULL;

  return g_hash_table_lookup(tcpc->data, key);
}


void           *
libnd_tcpconn_data_remove(LND_TCPConn *tcpc,
			  const char *key)
{
  void *data;

  if (!tcpc || !key || !*key)
    return NULL;
  
  data = g_hash_table_lookup(tcpc->data, key);
  g_hash_table_remove(tcpc->data, key);

  return data;
}


GHashTable     *
libnd_tcpconn_table_new(void)
{
  return g_hash_table_new((GHashFunc) tcp_conn_hash,
			  (GCompareFunc) tcp_conn_compare);
}


void            
libnd_tcpconn_table_free(GHashTable *ht)
{
  if (ht)
    {
      g_hash_table_foreach_remove(ht, (GHRFunc) libnd_tcpconn_free, NULL);
      g_hash_table_destroy(ht);
    }
}


void
libnd_tcpconn_table_add(GHashTable *ht, LND_TCPConn *tcpc)
{
  if (!ht || !tcpc)
    return;

  g_hash_table_insert(ht, (gpointer) tcpc, (gpointer) tcpc);
}


LND_TCPConn    *
libnd_tcpconn_table_lookup(GHashTable *ht,
			   const LND_Packet *packet)
{
  struct ip *iphdr;
  struct tcphdr *tcphdr;
  LND_TCPConn conn;
  LND_TCPConn *result;
  struct bpf_timeval diff;

  if (!ht || !packet)
    return NULL;
  if (! libnd_tcp_get_headers(packet, &iphdr, &tcphdr))
    return NULL;

  memset(&conn, 0, sizeof(LND_TCPConn));
  conn.ip_src   = iphdr->ip_src;
  conn.ip_dst   = iphdr->ip_dst;
  conn.th_sport = tcphdr->th_sport; 
  conn.th_dport = tcphdr->th_dport; 

  if (! (result = g_hash_table_lookup(ht, (gconstpointer) &conn)))
    return NULL;

  /* We need to check the various possible timeouts for
   * expiration. If any timeouts expire, we update the
   * existing connection to its terminated state, and
   * report no found connection. The user then creates
   * a new one, and the old one is never matched in the
   * hash table because we defined the comparison operator
   * to always reject closed connections.
   */
  if (result->state == LND_TCP_SYN_SENT)
    {
      pcapnav_timeval_sub(&packet->ph.ts, &result->timeout_ts, &diff);
      
      if (diff.tv_sec >= LND_TCP_TIMEOUT)
	{
	  result->state = LND_TCP_CLOSED_TIMEOUT;
	  return NULL;
	}
    }
  
  if (result->state == LND_TCP_SHUTDOWN)
    {
      pcapnav_timeval_sub(&packet->ph.ts, &result->timeout_ts, &diff);
      
      if (diff.tv_sec >= LND_TCP_TIMEOUT)
	{
	  result->state = LND_TCP_CLOSED_TIMEOUT;
	  return NULL;
	}
    }
  
  if (result->state == LND_TCP_TIME_WAIT)
    {  
      pcapnav_timeval_sub(&packet->ph.ts, &result->wait_ts, &diff);
      
      if (diff.tv_sec >= 2 * LND_TCP_MSL)
	{
	  result->state = LND_TCP_CLOSED_NORMAL;
	  return NULL;
	}
    }
  
  if (result->state == LND_TCP_RST_WAIT)
    {	  
      pcapnav_timeval_sub(&packet->ph.ts, &result->wait_ts, &diff);
      
      if (diff.tv_sec >= LND_TCP_MSL)
	{
	  result->state = LND_TCP_CLOSED_RST;
	  return NULL;
	}
    }
  
  return result;
}


void
libnd_tcpconn_table_remove(GHashTable *ht, LND_TCPConn *conn)
{
  g_hash_table_remove(ht, conn);
}


struct tcp_conn_table_cb
{
  LND_TCPConnFunc  func;
  void            *user_data;
};

static void
tcpconn_table_foreach_cb(gpointer key, gpointer val, void *user_data)
{
  struct tcp_conn_table_cb *cb_data = user_data;
  cb_data->func((LND_TCPConn *)val, cb_data->user_data);

  return;
  TOUCH(key);
}

void
libnd_tcpconn_table_foreach(GHashTable *ht,
			    LND_TCPConnFunc func,
			    void *user_data)
{
  struct tcp_conn_table_cb cb_data;

  if (!ht || !func)
    return;

  cb_data.func = func;
  cb_data.user_data = user_data;

  g_hash_table_foreach(ht, tcpconn_table_foreach_cb, &cb_data);
}


const char *
name(void)
{
  return "TCP-State-Tracker";
}

const char *
author(void)
{
  return "Christian Kreibich, <christian.kreibich@cl.cam.ac.uk>";
}

const char *
version(void)
{
  return VERSION;
}

/* No global initialization necessary -> No init(). */


static void
tcpstate_help(void)
{
  printf("TCP Connection State Tracker plugin\n"
	 "USAGE: lndtool -r tcp-state-tracker <trace1> [<trace2> <trace3> ...]\n"
	 "\n"
	 "  --help, -h, -?           This message.\n");
}

gboolean
run(LND_Trace *unused, LND_PluginArgs *args)
{
  int i;
  LND_Trace *trace;
  LND_PacketIterator pit;
  LND_Packet *packet;
  LND_TCPConn *conn;
  GHashTable *ht;
  char line[MAXPATHLEN];
  gboolean new_conn;

  if (!args)
    return FALSE;

  if (args->argc == 0)
    {
      printf("Please provide one or more traces to scan.\n");
      tcpstate_help();
      return FALSE;
    }

  for (i = 0; i < args->argc; i++)
    {
      if (!strcmp(args->argv[i], "-h")     ||
	  !strcmp(args->argv[i], "--help") ||
	  !strcmp(args->argv[i], "-?"))
	{
	  tcpfilter_test_help();
	  return FALSE;
	}
      else
	{
	  printf("Scanning file %s\n", args->argv[i]);
	  
	  if (! (trace = libnd_trace_new(args->argv[i])))
	    {
	      printf("Could not open %s, skipping.\n", args->argv[i]);
	      continue;
	    }
	  
	  ht = libnd_tcpconn_table_new();
	  
	  for (libnd_pit_init(&pit, trace); libnd_pit_get(&pit); libnd_pit_next(&pit))
	    {
	      packet = libnd_pit_get(&pit);
	      libnd_tcpdump_get_packet_line(packet, line, FALSE);
	      new_conn = FALSE;
	      
	      if (! (conn = libnd_tcpconn_table_lookup(ht, packet)))
		{
		  new_conn = TRUE;
		  conn = libnd_tcpconn_new(packet);
		  libnd_tcpconn_table_add(ht, conn);
		}
	      
	      libnd_tcpconn_state_update(conn, packet);
	      printf("%s [%s%s]\n", line,
		     libnd_tcpconn_state_get_string(libnd_tcpconn_state_get(conn)),
		     new_conn ? "*" : "");
	    }
	  
	  libnd_tcpconn_table_free(ht);
	  libnd_trace_free(trace);
	}
    }
  
  return TRUE;
  TOUCH(unused);
}

